from cryptography.fernet import Fernet
from config.apps import FERNET_KEY


class Cookie:
    def __init__(self, request):
        self.request = request

    def set_server_cookie(self, header):
        """
        Set the cookies that the user writes to the request.
        After the data is written to the header the field in the request is cleared.
        """
        if self.request.get_user_set_cookie():
            header.extend(("set-cookie", morsel.OutputString())
                          for morsel in self.request.get_user_set_cookie().values())
            # clear request cookies
            self.request.clear_request_cookie()
            return header

    @classmethod
    def encrypt_cookie(cls, value: str):
        key = FERNET_KEY
        fernet = Fernet(key)
        return fernet.encrypt(value.encode())

    @classmethod
    def decrypt_cookie(cls, value: bytes):
        key = FERNET_KEY
        fernet = Fernet(key)
        return fernet.decrypt(value).decode()



import fire
from flexypy.generate_templates.generator import generate


class Command:
    @staticmethod
    def initproject(dir_path):
        generate(dir_path)


def init_commands():
    fire.Fire(Command())



from flexypy.http.request import Request
from dataclasses import dataclass
import re
from flexypy.exceptions.routing import PathParameterAlreadyExists
from abc import abstractmethod
from flexypy.http.template.render import RenderTemplate
from typing import Callable


@dataclass
class SlugData:
    count: int
    params: dict


class Route:
    def __init__(self, path, template_path):
        self.path = path
        self.template_path = template_path
        self.method = ''
        self.get_func = None
        self.post_func = None
        self.request: Request = None
        self.template_variables = {}
        self.template_methods = []

    def set_template_methods(self, functions: list[Callable]):
        self.template_methods = functions

    def get_path(self):
        return self.path

    def get(self):
        if self.get_func:
            self.get_func()
        return RenderTemplate(self.template_path).set_template_methods(self.template_methods)\
            .render(**self.template_variables)

    def set_template_variables(self, **kwargs):
        self.template_variables = kwargs

    @abstractmethod
    def post(self):
        pass

    def set_get(self, func):
        self.get_func = func

    def set_post(self, func):
        self.post_func = func

    def parse_slug(self, path) -> SlugData:
        names = self._get_slug_names(path)
        count = len(names)-1
        params = {}
        for name in names:
            for n in re.finditer(name, path):
                params[f"[{name}]"] = [n.start()-1, n.end()+1]
        return SlugData(count, params)

    def _get_slug_names(self, path):
        names = []
        if path.find('[') != -1 and path.find(']') != -1:
            end = path.find(']')+1
            name = path[path.find('[')+1:path.find(']'):]
            names.append(name)
            p = path[end::]
            if p.find('[') != -1:
                for i in self._get_slug_names(p):
                    if i in names:
                        raise PathParameterAlreadyExists('_get_slug_names', i)
                    names.append(i)
        return names


class UserRoute(Route):
    def __init__(self, path, template_path, parent_route=None):
        self.parent_route = parent_route
        p = self._modify_path(path)
        self.slug_data = self.parse_slug(p)
        super().__init__(p, template_path)

    def _modify_path(self, path) -> str:
        if self.parent_route:
            return self.parent_route().path.strip('/') + '/' + path.lstrip('/')
        else:
            return path

    def post(self):
        pass


def redirect(path):
    return path




from flexypy.http.request import Request
from dataclasses import dataclass
import mimetypes
from config.apps import PATH_TO_FXP_APPS
from config.dirs import MDDL_LIST_PATH
import importlib
from wsgiref import util
from flexypy.http.routing import UserRoute
from typing import Type
import os
from flexypy.exceptions.web.server import PathNotFound
from urllib import parse
from cgi import FieldStorage
from flexypy.middlewares.mddl import Middleware
import re
from flexypy.http.cookie import Cookie


@dataclass
class HtmlResponse:
    code: str
    header: list[tuple[str, str]]
    html: bytes


@dataclass
class MddlRedirect:
    from_path: str
    to_path: str


class WsgiServer:
    current_app_path = ''

    def __init__(self, environ: dict, start_response):
        self.environ = environ
        self.start_response = start_response
        self.root_url = util.shift_path_info(self.environ)
        self.full_url = self.root_url + util.request_uri(self.environ).replace(util.application_uri(self.environ), '')
        self.server_address = util.application_uri(self.environ).replace(self.root_url, '')
        self.render: RenderTemplate = RenderTemplate()
        self.routes: list[Type[UserRoute]] = []
        self.router = None
        self.request = Request
        self._get_fxp_apps()
        self.router = Router(self.environ, self.routes, self.full_url)
        self._set_server_cookie()

    def _set_server_cookie(self):
        try:
            if 'HTTP_COOKIE' in self.environ:
                self.request.set_server_cookie(self.environ['HTTP_COOKIE'])
        except Exception as e:
            pass

    def _route(self) -> HtmlResponse:
        path_found = False
        match self.environ['REQUEST_METHOD']:
            case 'GET':
                if self.router.check_url():
                    path_found = True
                    app = self.router.check_url()()

                    # Set request data
                    self.request.GET.set_data(self.router.query_kwargs)
                    self.request.path = self.full_url

                    # Set app request
                    app.request = self.request

                    mddl_app, mddl_redirect = self._run_middlewares(app)
                    if mddl_redirect:
                        if mddl_redirect.from_path and mddl_redirect.to_path:
                            if self.full_url == mddl_redirect.from_path:
                                return self.render.render_redirect(mddl_redirect.to_path)
                    return self._method_get(mddl_app)

                if self.router.check_static_file():
                    return self._method_get_static_files(self.router.check_static_file())

                # If path not found
                if not path_found:
                    t = PathNotFound(self.full_url, [self.server_address + i().get_path() for i in self.routes])
                    return self.render.render_traceback(t.code, t.html)
            case 'POST':
                if self.router.check_url():
                    app = self.router.check_url()()

                    # Set request data
                    self.request.GET.params = self.router.query_kwargs
                    form_data = FieldStorage(fp=self.environ['wsgi.input'], environ=self.environ,
                                             keep_blank_values=True)
                    self.request.POST.set_data(form_data)

                    # Set app request
                    app.request = self.request

                    mddl_app, mddl_redirect = self._run_middlewares(app)
                    if mddl_redirect:
                        if mddl_redirect.from_path and mddl_redirect.to_path:
                            if self.full_url == mddl_redirect.from_path:
                                return self.render.render_redirect(mddl_redirect.to_path)

                    return self._method_post(mddl_app)

    def _method_get(self, app: UserRoute) -> HtmlResponse:
        mime_type = mimetypes.guess_type(app.template_path)[0]
        resp = self.render.render_html('200 OK', [('Content-type', mime_type)], app.get())
        return resp

    def _method_post(self, app: UserRoute):
        resp = self.render.render_redirect(app.post())
        return resp

    def _method_get_static_files(self, filepath) -> HtmlResponse:
        # get static filepath
        if 'HTTP_REFERER' in self.environ:
            current_url: str = self.environ['HTTP_REFERER'].replace(self.server_address, '')
            if current_url.endswith('/'):
                p = filepath.split(current_url)[-1].strip('/')
            elif not current_url:
                p = filepath
            else:
                if current_url.rfind('/') != -1:
                    converted_path = current_url[:current_url.rfind('/')+1:]
                else:
                    converted_path = current_url

                p = filepath

                if filepath.startswith(converted_path):
                    p = filepath.split(converted_path)[-1].strip('/')

                    # If names are duplicated in the path. For example test/test/path/to/file
                    for i in re.finditer(converted_path, filepath):
                        if os.path.exists(filepath[i.end()::]):
                            p = filepath[i.end()::]
                            break

            if os.path.exists(p):
                with open(p, 'rb') as f:
                    mime_type = mimetypes.guess_type(p)[0]
                    resp = self.render.render_html('200 OK', [('Content-type', mime_type)], f.read())
                    return resp

    def _run_middlewares(self, app: UserRoute) -> list[UserRoute, MddlRedirect]:
        mddl = (os.path.join(MDDL_LIST_PATH, 'mddl_list'))
        mddl_app = app
        redirect = None
        if os.path.exists(mddl+'.py'):
            global_t_obj = __import__(mddl.replace('/', '.'), fromlist=['middlewares'])
            mddl_list = getattr(global_t_obj, 'middlewares')

            for mddl in mddl_list:
                m: Middleware = mddl(self.request, app)
                m.start()
                mddl_app = m.app
                redirect = MddlRedirect(m.redirect_from, m.redirect_to)
        return [mddl_app, redirect]

    def start(self):
        resp = self._route()

        if 'HTTP_COOKIE' in self.environ:
            self.request.set_server_cookie(self.environ['HTTP_COOKIE'])
        if resp:
            # set header cookies
            header = Cookie(self.request).set_server_cookie(resp.header)
            if header:
                resp.header = header

            self.start_response(resp.code, resp.header)
            return [resp.html]
        else:
            mime_type = self.environ['SCRIPT_NAME']
            header = [('Content-type', mime_type)]
            self.start_response('404 Not found', header)
            return [b'']

    def _get_fxp_apps(self):
        for p in PATH_TO_FXP_APPS:
            import_path = p.replace('/', '.').strip('.') + '.fxp'
            importlib.import_module(import_path)
        self.routes = UserRoute.__subclasses__()


class RenderTemplate:
    def __init__(self):
        pass

    def _text_to_html(self, text: str | bytes) -> bytes:
        if isinstance(text, bytes):
            return bytes(text)
        elif isinstance(text, str):
            return bytes(text, 'utf-8')

    def render_html(self, code, header: list[tuple[str, str]], html: str | bytes) -> HtmlResponse:
        h = self._text_to_html(html)
        return HtmlResponse(code, header, h)

    def render_404(self, html: str | bytes):
        return HtmlResponse('404 Not Found', [('Content-type', 'text/html')], self._text_to_html(html))

    def render_traceback(self, code: str, html: str | bytes):
        return HtmlResponse(code, [('Content-type', 'text/html')], self._text_to_html(html))

    def render_redirect(self, path):
        return HtmlResponse('303 See Other', [('Location', path)], b'')


class Router:
    def __init__(self, environ,  apps: list[Type[UserRoute]], current_url: str):
        self.environ = environ
        self.apps = apps
        self.current_url = current_url
        self.query_kwargs = {}

    def check_url(self) -> Type[UserRoute] | None:
        url, self.query_kwargs = self._parse_qs(self.current_url)

        for app in self.apps:
            a = app()
            check_path = a.path
            url_params = {}
            # Check slug fields
            if a.slug_data.count != -1:
                url_params = self._get_url_params(a)
                check_path = a.path
                for p in url_params:
                    check_path = check_path.replace(f"[{p}]", url_params[p][0])

            if check_path == url:
                self.query_kwargs.update(url_params)
                return app
        return None

    def _get_url_params(self, app) -> dict:
        par = {}
        u = self.current_url
        for sp in app.slug_data.params:
            p = u[app.slug_data.params[sp][0]::]
            if p.find('/') != -1:
                u = app.path[:app.slug_data.params[sp][1]:] + p[p.find('/')::]
                p = p[:p.find('/'):]
            if p:
                par[sp.strip('[').strip(']')] = [p]
            else:
                return par
        return par

    def _parse_qs(self, path: str) -> list[str, dict]:
        p = path.split('?')
        return [p[0], parse.parse_qs(self.environ['QUERY_STRING'])]

    def check_static_file(self):
        if self.current_url.rfind('.') != -1:
            return self.current_url
        else:
            return None



from flexypy.generate_templates.templates.config import dirs, apps
from flexypy.generate_templates.templates import main
import os


files = [
    {'dir': 'config', 'name': 'dirs.py', 'path': dirs.__file__},
    {'dir': 'config', 'name': 'apps.py', 'path': apps.__file__},
    {'dir': '', 'name': 'main.py', 'path': main.__file__},
]


def generate(generate_dir):

    if generate_dir == '.':
        gd = os.getcwd()
    elif generate_dir == '':
        raise 'ERR'
    else:
        gd = generate_dir

    for file in files:
        dir_path = os.path.join(gd, file['dir'])
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)

        if os.path.exists(dir_path):
            if not os.path.exists(os.path.join(dir_path, file['name'])) and file['name']:
                with open(os.path.join(dir_path, file['name']), 'w+') as wf:
                    with open(file['path'], 'r') as rf:
                        wf.write(rf.read())


if __name__ == '__main__':
    generate('/home/uwine/Documents/python/flexypy/')



from .baseexceptions import FlexypyBaseException


class PathParameterAlreadyExists(FlexypyBaseException):
    def __init__(self, func_name, parameter_name):
        super().__init__(func_name=func_name, text=f'Path parameter "{parameter_name}" already exists.')



from flexypy.exceptions.web.baseexseptions import BaseWebException


class PathNotFound(BaseWebException):
    def __init__(self, not_found_path, paths: list[str]):
        self.paths = paths
        self.not_found_path = not_found_path
        super().__init__('404 Not found', self.get_html_text())

    def get_html_text(self):
        p = ''
        for path in self.paths:
            p += f"<p>{path}</p>"
        text = f"""
            <h1>Path "{self.not_found_path}" not found</h1>
            <h2>Accept paths:</h2>
            {p}        
        """
        return text




class Middleware:
    def __init__(self, request, app):
        self.request = request
        self.app = app
        self.redirect_from = ''
        self.redirect_to = ''

    def start(self):
        pass

    def redirect(self, from_path, to_path):
        self.redirect_from = from_path
        self.redirect_to = to_path



from .baseexceptions import FlexypyBaseException


class TemplateNotFound(FlexypyBaseException):
    def __init__(self, func_name, template_path):
        super().__init__(func_name=func_name, text=f'Template {template_path} not found.')




from jinja2.ext import Extension
from jinja2 import nodes
from abc import abstractmethod


class BaseExtension(Extension):
    """
    Base extension class. All extensions should inherit from it.
    """

    # Extensiom name.
    tags = []

    def parse(self, parser):
        lineno = parser.stream.expect(f'name:{self.tags[0]}').lineno
        parse_expression = [parser.parse_expression()]
        if parser.stream.skip_if("comma"):
            parse_expression.append(parser.parse_expression())
        call = self.call_method('handler', parse_expression, lineno=lineno)
        return nodes.Output([nodes.MarkSafe(call)]).set_lineno(lineno)

    @abstractmethod
    def handler(self) -> str:
        """
        Extension handler.
        """
        pass


from flexypy.http.template.extensions.base_extension import BaseExtension
from config.dirs import STATICFILES_PATH
import os
from flexypy.exceptions.extension import StaticFileNotFound


class StaticExt(BaseExtension):
    """
    Extensions for outputting files to html template.
    """
    tags = ['static']

    def handler(self, filepath: str):
        for path in STATICFILES_PATH:
            p = os.path.join(path, filepath)
            if os.path.exists(p):
                return p
        raise StaticFileNotFound('StaticExt', filepath)



import jinja2
from config.dirs import TEMPLATES_PATH, EXTENSIONS_PATH, GLOBAL_TEMPLATE_OBJECTS_PATH, GLOBAL_TEMPLATE_FUNCTIONS_PATH
import os
from typing import Callable, Type
from flexypy.http.template.extensions.base_extension import BaseExtension
import importlib
from flexypy.exceptions.render import TemplateNotFound


class RenderTemplate:
    def __init__(self, template_path):
        self.template_path = template_path
        self._template_source: str = ''
        self._methods = []

    def set_template_methods(self, functions: list[Callable]):
        for fun in functions:
            self._methods.append(fun)
        return self

    def render(self, **kwargs) -> str:
        try:
            for templ in TEMPLATES_PATH:
                filepath = os.path.join(templ, self.template_path)
                if os.path.exists(filepath):
                    return _JinjaInit(filepath).template_methods(self._methods).jinit(**kwargs)
            raise TemplateNotFound('render', self.template_path)
        except Exception as e:
            raise e


class _JinjaInit:
    def __init__(self, filepath: str):
        self._path = filepath.split('templates')[-1]
        self._env = jinja2.Environment(loader=self._get_loader(), extensions=self._get_extensons())
        self._template = self._env.get_template(self._path)
        self._template_obj = []

    def template_methods(self, methods: list[callable]):
        """
        Adding Python methods to all templates.
        """
        for meth in methods:
            methname = meth.__name__
            self._env.globals[methname] = meth
        return self

    def _get_loader(self) -> jinja2.ChoiceLoader:
        """
        Creates loaders for html templates.
        """
        try:
            fsl = []
            for template in TEMPLATES_PATH:
                fsl.append(jinja2.FileSystemLoader(template))
            loader = jinja2.ChoiceLoader(fsl)
            return loader
        except Exception as e:
            raise e

    def _get_extensons(self) -> list[Type[BaseExtension]]:
        """
        Collects all template extensions (Classes) inherited from BaseExtension.
        """

        # dafault extensions path
        EXTENSIONS_PATH.append('flexypy/http/template/extensions/')

        for path in EXTENSIONS_PATH:
            p = path.replace('/', '.')
            importlib.import_module(f'{p}template_ext')
        return BaseExtension.__subclasses__()

    def jinit(self, **kwargs):
        objects = kwargs

        # set global methods
        global_objects = (os.path.join(GLOBAL_TEMPLATE_OBJECTS_PATH, 'global_objects'))
        if os.path.exists(global_objects+'.py'):
            global_t_obj = __import__(global_objects.replace('/', '.'), fromlist=['objects'])
            objects = getattr(global_t_obj, 'objects')
            objects.update(kwargs)

        # set global functions
        global_functions = os.path.join(GLOBAL_TEMPLATE_FUNCTIONS_PATH, 'global_functions')
        if os.path.exists(global_functions+'.py'):
            global_t_obj = __import__(global_functions.replace('/', '.'), fromlist=['objects'])
            self.template_methods(getattr(global_t_obj, 'objects'))

        for obj in self._template_obj:
            kwargs[obj] = self._template_obj[obj]
        render = self._template.render(objects)
        return render



from cgi import FieldStorage
from abc import abstractmethod
from http import cookies
from flexypy.http.cookie import Cookie
import datetime


class FormData:
    @abstractmethod
    def get(cls, key) -> list:
        pass

    @abstractmethod
    def keys(cls) -> list:
        pass

    @abstractmethod
    def get_data(cls) -> FieldStorage:
        pass


class PostFormData(FormData):
    _form_data: FieldStorage

    @classmethod
    def get(cls, key) -> list:
        return cls._form_data.getlist(key)

    @classmethod
    def keys(cls) -> list:
        return cls._form_data.keys()

    @classmethod
    def get_data(cls) -> FieldStorage:
        return cls._form_data

    @classmethod
    def set_data(cls, form_data: FieldStorage):
        cls._form_data = form_data


class GetFormData(FormData):
    _form_data: dict

    @classmethod
    def get(cls, key):
        return cls._form_data.get(key)

    @classmethod
    def keys(cls) -> list:
        return list(cls._form_data.keys())

    @classmethod
    def get_data(cls) -> dict:
        return cls._form_data

    @classmethod
    def set_data(cls, form_data: dict):
        cls._form_data = form_data


class Request:
    _set_cookie: cookies.SimpleCookie = {}
    _server_cookie: cookies.SimpleCookie = None
    path = ''
    POST = PostFormData
    GET = GetFormData

    @classmethod
    def get_user_set_cookie(cls):
        return cls._set_cookie

    @classmethod
    def clear_request_cookie(cls):
        cls._set_cookie = {}

    @classmethod
    def set_cookie(cls, name, value, expires=None):
        c = cookies.SimpleCookie()
        c[name] = Cookie.encrypt_cookie(value).decode()
        c[name]['expires'] = (datetime.datetime.now() + datetime.timedelta(days=365)). \
            strftime('%a, %d-%b-%Y %H:%M:%S GMT')
        c[name]['secure'] = True
        if expires:
            c[name]['expires'] = expires
        cls._set_cookie.update(c)

    @classmethod
    def set_server_cookie(cls, cookie):
        c = cookies.SimpleCookie()
        c.load(cookie)
        cls._server_cookie = c

    @classmethod
    def get_server_cookie(cls, name):
        try:
            return Cookie.decrypt_cookie(cls._server_cookie.get(name).value.encode())
        except:
            return None

    @classmethod
    def delete_cookie(cls, name):
        cls.set_cookie(name, '', expires='Thu, 01, Jan 1970 00:00:00 GMT')



from .baseexceptions import FlexypyBaseException


class StaticFileNotFound(FlexypyBaseException):
    def __init__(self, func_name, static_file_name):
        super().__init__(func_name=func_name, text=f'Static file "{static_file_name}" not found.')


class FlexypyBaseException(Exception):
    def __init__(self, func_name: str, text: str):
        self.text = text
        self.func_name = func_name

    def __str__(self):
        return f'{self.func_name}: {self.text}'



TEMPLATES_PATH = [
    'templates',
]

STATICFILES_PATH = [
    'static',
]

EXTENSIONS_PATH = [

]

GLOBAL_TEMPLATE_OBJECTS_PATH = 'global'
GLOBAL_TEMPLATE_FUNCTIONS_PATH = 'global'
MDDL_LIST_PATH = 'global'
